/**
  *  \file game/actions/basefix.cc
  *  \brief Starbase Action: Fix/Recycle
  *
  *  (c) 2001-2015 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/actions/basefix.h"
#include "util/nls.h"
#include "game/univ.h"
#include "game/objlist.h"
#include "game/ship.h"
#include "game/global.h"
#include "game/actions/preconditions.h"

/*********************** GStarbaseFixRecycleAction ***********************/

/** \class GStarbaseFixRecycleAction
    \brief Starbase Action: Fix/Recycle.

    This class encapsulates all rules for a Fix/Recycle action. */

/** Initialize. */
GStarbaseFixRecycleAction::GStarbaseFixRecycleAction(GUniverse& univ, GPlanet& planet)
    : univ(univ), planet(planet)
{
    mustHavePlayedBase(planet);
}

/** Set shipyard action.
    \param action Action to perform
    \param ship_id Ship to work on; 0 if action is sya_None
    \return true on success, false if order not accepted. */
bool
GStarbaseFixRecycleAction::setAction(TShipyardAction action, int ship_id)
{
    if (action == sya_None) {
        // None -> ignore action
        ship_id = 0;
    } else {
        // Verify action
        ASSERT(action == sya_Recycle || action == sya_Fix);
        if (!univ.ty_any_ships.isValidIndex(ship_id)) {
            return false;
        }

        if ((getValidBaseActionsForShip(ship_id) & (1 << action)) == 0) {
            return false;
        }
    }
    planet.setBaseShipyardOrder(mp16_t(action), mp16_t(ship_id));
    return true;
}

/** Get permitted starbase actions.
    \return Bitfield of CanRecycleFlag, CanFixFlag. */
int
GStarbaseFixRecycleAction::getValidBaseActions() const
{
    int flags = 0;
    for (int sid = univ.ty_any_ships.findNextIndex(0); sid != 0; sid = univ.ty_any_ships.findNextIndex(sid)) {
        flags |= getValidBaseActionsForShip(sid);
    }
    return flags;
}

/** Enumerate ships for a particular action.
    \param action desired action (sya_Fix, sya_Recycle).
    \param list   output list, should be empty. */
void
GStarbaseFixRecycleAction::enumerateShipsFor(TShipyardAction action, GObjectList& list) const
{
    if (action != sya_None) {
        ASSERT(action == sya_Recycle || action == sya_Fix);
        for (int sid = univ.ty_any_ships.findNextIndex(0); sid != 0; sid = univ.ty_any_ships.findNextIndex(sid)) {
            if ((getValidBaseActionsForShip(sid) & (1 << action)) != 0) {
                list.addObject(univ.ty_any_ships, sid);
            }
        }
    }
}

/** Get actions that are valid for this ship.
    Returns 0 if the ship is at the wrong position or does not exist.
    Otherwise, returns a bitfield of
    - CanFixFlag: we can fix it
    - CanRecycleFlag: we can recycle it
    - SamePositionFlag: we cannot do anything with it, but it's at the same position */
int
GStarbaseFixRecycleAction::getValidBaseActionsForShip(int ship_id) const
{
    if (!univ.ty_any_ships.isValidIndex(ship_id)) {
        // Does not exist
        return 0;
    } else {
        const GShip& sh = univ.getShip(ship_id);
        if (sh.getPos() != planet.getPos()) {
            // Wrong position
            return 0;
        } else if (sh.getOwner() == planet.getOwner()) {
            // Same owner: can do everything
            return CanFixFlag | CanRecycleFlag;
        } else {
            // Different owner.
            // FIXME: this should have actual host rules.
            // PHost: allows Fix to allies
            // THost: allows everything to everyone up to a certain version, only to owner from then
            return CanFixFlag;
        }
    }
}

/** Check a base action.
    For use by the script interface.

    FIXME: let this directly set the action. */
GError
checkBaseAction(TShipyardAction action, GUniverse& univ, int ship_id, int planet_id)
{
    // Verify planet
    if (!univ.isValidPlanetId(planet_id))
        return GError(GError::eRange, GError::eRange);
    GStarbaseFixRecycleAction act(univ, univ.getPlanet(planet_id));

    // Verify ship
    int flag = act.getValidBaseActionsForShip(ship_id);
    if ((flag & (1 << action)) != 0) {
        // Success
        return GError();
    } else if (flag == 0) {
        return GError(GError::ePos, GError::ePos);
    } else {
        return GError(GError::ePerm, GError::ePerm);
    }
}
